@@ -17,6 +17,7 @@ char8 c = () <$ single (fromIntegral (ord c))
 satisfy8 :: (Char -> Bool) -> Parser ()
 satisfy8 f = () <$ satisfy (f . chr . fromIntegral)
 
+ws, open, close, ident, sexp :: Parser ()
 ws      = skipMany (satisfy8 \c -> c == ' ' || c == '\n')
 open    = char8 '(' >> ws
 close   = char8 ')' >> ws
@@ -24,10 +25,12 @@ ident   = skipSome (satisfy8 isLatinLetter) <* ws
 sexp    = (open *> skipSome sexp <* close) <|> ident
 runSexp = runParser sexp ""
 
-longw     = chunk "thisisalongkeyword"
+longw, longws :: Parser ()
+longw     = () <$ chunk "thisisalongkeyword"
 longws    = skipSome (longw *> ws) <* eof
 runLongws = runParser longws ""
 
+numeral, comma, numcsv :: Parser ()
 numeral   = skipSome (satisfy8 \c -> '0' <= c && c <= '9') >> ws
 comma     = single (fromIntegral (ord ',')) >> ws
 numcsv    = numeral >> skipSome (comma >> numeral) >> eof
