@@ -3,6 +3,7 @@
 {-# LANGUAGE KindSignatures #-}
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 
 {-|
 Parser supporting custom error types and embeddable `IO` or `ST` actions, but
@@ -22,6 +23,8 @@ module FlatParse.Basic (
   , runParserUtf8
   , runParserIO
   , runParserST
+  , embedParserST
+  , embedParser
 
   -- ** Primitive result types
   , type FP.Parser.Res#
@@ -73,6 +76,7 @@ module FlatParse.Basic (
   , FP.Base.withEnsure1
   , FP.Base.withEnsure#
   , FP.Base.isolate
+  , isolateToNextNull
   , FP.Base.isolate#
   , FP.Base.isolateUnsafe#
   , FP.Switch.switch
@@ -88,6 +92,8 @@ module FlatParse.Basic (
   , FP.Base.failed
   , FP.Base.try
   , FP.Base.err
+  , FP.Base.withError
+  , FP.Base.withAnyResult
   , FP.Base.fails
   , FP.Base.cut
   , FP.Base.cutting
@@ -153,6 +159,9 @@ module FlatParse.Basic (
 
   ) where
 
+-- for WORDS_BIGENDIAN
+#include "MachDeps.h"
+
 import FlatParse.Basic.Parser
 import FlatParse.Basic.Base
 import FlatParse.Basic.Integers
@@ -174,7 +183,7 @@ import qualified FlatParse.Basic.Switch as FP.Switch
 import qualified FlatParse.Basic.Addr as FP.Addr
 
 import qualified Control.Applicative
-import GHC.IO (IO(..))
+import GHC.IO (IO(..), unsafeIOToST)
 import GHC.Exts
 import GHC.ForeignPtr
 import GHC.ST (ST(..))
@@ -211,7 +220,7 @@ unsafeLiftIO io = ParserT \fp eob s st ->
 
 --------------------------------------------------------------------------------
 
--- | Run a parser.
+-- | Run a pure parser.
 runParser :: Parser e a -> B.ByteString -> Result e a
 runParser (ParserT f) b@(B.PS (ForeignPtr _ fp) _ (I# len)) = unsafePerformIO $
   B.unsafeUseAsCString b \(Ptr buf) -> do
@@ -226,7 +235,7 @@ runParser (ParserT f) b@(B.PS (ForeignPtr _ fp) _ (I# len)) = unsafePerformIO $
 -- We mark this as noinline to allow power users to safely do unsafe state token coercions.
 -- Details are discussed in https://github.com/AndrasKovacs/flatparse/pull/34#issuecomment-1326999390
 
--- | Run a parser on a 'String', converting it to the corresponding UTF-8 bytes.
+-- | Run a pure parser on a 'String', converting it to the corresponding UTF-8 bytes.
 --
 -- Reminder: @OverloadedStrings@ for 'B.ByteString' does not yield a valid UTF-8
 -- encoding! For non-ASCII 'B.ByteString' literal input, use this wrapper or
@@ -235,8 +244,8 @@ runParserUtf8 :: Parser e a -> String -> Result e a
 runParserUtf8 pa s = runParser pa (Common.strToUtf8 s)
 
 -- | Run an `ST`-based parser.
-runParserST :: (forall s. ParserST s e a) -> B.ByteString -> Result e a
-runParserST pst buf = unsafeDupablePerformIO (runParserIO pst buf)
+runParserST :: ParserST s e a -> B.ByteString -> ST s (Result e a)
+runParserST pst buf = unsafeIOToST (runParserIO (unsafeCoerce# pst) buf)
 {-# inlinable runParserST #-}
 
 -- | Run an `IO`-based parser.
@@ -258,6 +267,16 @@ liftST (ST f) = ParserT \fp eob s st -> case f st of
   (# st, a #) -> OK# st a s
 {-# inline liftST #-}
 
+-- | Run a `ParserST` inside any parser.
+embedParserST :: forall e a s. (forall s. ParserST s e a) -> ParserT s e a
+embedParserST f = unsafeCoerce# (f :: ParserST RealWorld e a)
+{-# inline embedParserST #-}
+
+-- | Run a pure `Parser` inside any parser.
+embedParser :: forall e a s. Parser e a -> ParserT s e a
+embedParser f = unsafeCoerce# f
+{-# inline embedParser #-}
+
 --------------------------------------------------------------------------------
 
 -- | Parse a given 'B.ByteString'.
@@ -386,7 +405,7 @@ validPos str pos =
         _      -> error "FlatParse.Basic.validPos: got a non-OK result, impossible"
 {-# inline validPos #-}
 
--- | Compute corresponding line and column numbers for each `Pos` in a list,
+-- | Compute corresponding line and column numbers (both starting from 0) for each `Pos` in a list,
 --   assuming UTF8 encoding. Throw an error on invalid positions. Note:
 --   computing lines and columns may traverse the `B.ByteString`, but it
 --   traverses it only once regardless of the length of the position list.
@@ -405,7 +424,7 @@ posLineCols str poss =
             go line (col + 1) ((i, pos):poss)
 
       sorted :: [(Int, Pos)]
-      sorted = sortBy (\(_, i) (_, j) -> compare j i) (zip [0..] poss)
+      sorted = sortBy (\(_, i) (_, j) -> compare i j) (zip [0..] poss)
 
   in case runParser (go 0 0 sorted) str of
        OK res _ -> snd <$> sortBy (comparing fst) res
@@ -436,26 +455,81 @@ linesUtf8 str =
 
 --------------------------------------------------------------------------------
 
+-- | Isolate the given parser up to (excluding) the next null byte.
+--
+-- Like 'isolate', all isolated bytes must be consumed. The null byte is
+-- consumed afterwards.
+--
+-- Useful for defining parsers for null-terminated data.
+isolateToNextNull :: ParserT st e a -> ParserT st e a
+isolateToNextNull (ParserT p) = ParserT \fp eob s st -> go fp eob s st s
+  where
+    goP fp sNull s0 st =
+        case p fp sNull s0 st of
+          OK# st' a s' ->
+            case eqAddr# s' sNull of
+              1# -> -- consumed up to null: skip null, return
+                    OK#   st' a (sNull `plusAddr#` 1#)
+              _  -> Fail# st' -- didn't consume fully up to null: fail
+          x -> x
+
+    go8 fp eob s0 st s =
+        case eqAddr# eob s of
+          1# -> Fail# st -- end of input, no null: fail
+          _  ->
+            let s' = s `plusAddr#` 1# in
+#if MIN_VERSION_base(4,16,0)
+            -- below may be made clearer with ExtendedLiterals (GHC 9.8)
+            case eqWord8# (indexWord8OffAddr# s 0#) (wordToWord8# 0##) of
+#else
+            case eqWord# (indexWord8OffAddr# s 0#) 0## of
+#endif
+              1# -> goP fp s   s0 st    --     0x00: isolate, execute parser
+              _  -> go8 fp eob s0 st s' -- not 0x00: next please!
+
+    {- The "find first null byte" algorithms used here are adapted from
+       Hacker's Delight (2012) ch.6.
+
+    We read a word (8 bytes) at a time for efficiency. The internal algorithm
+    does byte indexing, thus endianness matters. We switch between indexing
+    algorithms depending on compile-time native endianness. (The code
+    surrounding the indexing is endian-independent, so we do this inline).
+    -}
+    go fp eob s0 st s =
+        let sWord = s `plusAddr#` 8# in
+        case gtAddr# sWord eob of
+          1# -> -- <  8 bytes of input: revert to scanning byte by byte
+            -- we _could_ operate on a word and simply ensure not to use the
+            -- out-of-bounds data, which would be faster, but the act of
+            -- reading could probably segfault
+            go8 fp eob s0 st s
+          _  -> -- >= 8 bytes of input: use efficient 8-byte scanning
+#if defined(WORDS_BIGENDIAN)
+            -- big-endian ("L->R"): find leftmost null byte
+            let !x@(I# x#) = Common.zbytel'intermediate (I# (indexIntOffAddr# s 0#)) in
+#else
+            -- little-endian ("R->L"): find rightmost null byte
+            let !x@(I# x#) = Common.zbyter'intermediate (I# (indexIntOffAddr# s 0#)) in
+#endif
+            case x# ==# 0# of
+              1# -> go fp eob s0 st sWord -- no 0x00 in next word
+              _  -> -- 0x00 somewhere in next word
+#if defined(WORDS_BIGENDIAN)
+                let !(I# nullIdx#) = Common.zbytel'toIdx x in
+#else
+                let !(I# nullIdx#) = Common.zbyter'toIdx x in
+                -- TO TEST BE ON LE: change above CPP to zbytel, uncomment below
+                -- let !(I# nullIdx#) = Common.zbytel'toIdx (I# (word2Int# (byteSwap# (int2Word# x#)))) in
+#endif
+                let sNull = s `plusAddr#` nullIdx# in
+                goP fp sNull s0 st
+{-# inline isolateToNextNull #-}
+
 -- | Read a null-terminated bytestring (a C-style string).
 --
 -- Consumes the null terminator.
 anyCString :: ParserT st e B.ByteString
-anyCString = ParserT go'
-  where
-    go' fp eob s0 st = go 0# s0
-      where
-        go n# s = case eqAddr# eob s of
-          1# -> Fail# st
-          _  ->
-            let s' = plusAddr# s 1#
-#if MIN_VERSION_base(4,16,0)
-            -- TODO below is a candidate for improving with ExtendedLiterals!
-            in  case eqWord8# (indexWord8OffAddr# s 0#) (wordToWord8# 0##) of
-#else
-            in  case eqWord# (indexWord8OffAddr# s 0#) 0## of
-#endif
-                  1# -> OK# st (B.PS (ForeignPtr s0 fp) 0 (I# n#)) s'
-                  _  -> go (n# +# 1#) s'
+anyCString = isolateToNextNull takeRest
 {-# inline anyCString #-}
 
 -- | Read a null-terminated bytestring (a C-style string), where the bytestring
