@@ -49,14 +49,14 @@ data Tm
 --   keyword.
 ident :: Parser Name
 ident = token $ byteStringOf $
-  spanned (identStartChar *> many_ identChar) (\_ span -> fails (isKeyword span))
+  withSpan (identStartChar *> skipMany identChar) (\_ span -> fails (isKeyword span))
 
 -- | Parse an identifier, throw a precise error on failure.
 ident' :: Parser Name
 ident' = ident `cut'` (Msg "identifier")
 
 digit :: Parser Int
-digit = (\c -> ord c - ord '0') <$> satisfyASCII isDigit
+digit = (\c -> ord c - ord '0') <$> satisfyAscii isDigit
 
 int :: Parser Int
 int = token do
