@@ -3,6 +3,7 @@
 {-# LANGUAGE KindSignatures #-}
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 
 {-|
 Parser supporting a custom reader environment, custom error types and an 'Int'
@@ -22,6 +23,8 @@ module FlatParse.Stateful (
   , runParserUtf8
   , runParserIO
   , runParserST
+  , embedParserST
+  , embedParser
 
   -- ** Primitive result types
   , type FP.Parser.Res#
@@ -82,6 +85,7 @@ module FlatParse.Stateful (
   , FP.Base.withEnsure1
   , FP.Base.withEnsure#
   , FP.Base.isolate
+  , isolateToNextNull
   , FP.Base.isolate#
   , FP.Base.isolateUnsafe#
   , FP.Switch.switch
@@ -97,6 +101,8 @@ module FlatParse.Stateful (
   , FP.Base.failed
   , FP.Base.try
   , FP.Base.err
+  , FP.Base.withError
+  , FP.Base.withAnyResult
   , FP.Base.fails
   , FP.Base.cut
   , FP.Base.cutting
@@ -162,6 +168,8 @@ module FlatParse.Stateful (
 
   ) where
 
+-- for WORDS_BIGENDIAN
+#include "MachDeps.h"
 
 import qualified FlatParse.Basic as Basic
 import FlatParse.Stateful.Parser
@@ -181,7 +189,7 @@ import qualified FlatParse.Stateful.Switch as FP.Switch
 import qualified FlatParse.Stateful.Addr as FP.Addr
 
 import qualified Control.Applicative
-import GHC.IO (IO(..))
+import GHC.IO (IO(..), unsafeIOToST)
 import GHC.Exts
 import GHC.ForeignPtr
 import GHC.ST (ST(..))
@@ -243,8 +251,8 @@ runParserUtf8 :: Parser r e a -> r -> Int -> String -> Result e a
 runParserUtf8 pa r !n s = runParser pa r n (Common.strToUtf8 s)
 
 -- | Run an `ST`-based parser. The `Int` argument is the initial state.
-runParserST :: (forall s. ParserST s r e a) -> r -> Int -> B.ByteString -> Result e a
-runParserST pst !r i buf = unsafeDupablePerformIO (runParserIO pst r i buf)
+runParserST :: ParserST s r e a -> r -> Int -> B.ByteString -> ST s (Result e a)
+runParserST pst !r i buf = unsafeIOToST (runParserIO (unsafeCoerce# pst) r i buf)
 {-# inlinable runParserST #-}
 
 -- | Run an `IO`-based parser. The `Int` argument is the initial state.
@@ -260,6 +268,16 @@ runParserIO (ParserT f) !r (I# n) b@(B.PS (ForeignPtr _ fp) _ (I# len)) = do
       Fail# rw'  ->  (# rw', Fail #)
 {-# inlinable runParserIO #-}
 
+-- | Run a `ParserST` inside any parser.
+embedParserST :: forall s r e a. (forall s. ParserST s r e a) -> ParserT s r e a
+embedParserST f = unsafeCoerce# (f :: ParserST RealWorld r e a)
+{-# inline embedParserST #-}
+
+-- | Run a pure `Parser` inside any parser.
+embedParser :: forall s r e a. Parser r e a -> ParserT s r e a
+embedParser f = unsafeCoerce# f
+{-# inline embedParser #-}
+
 --------------------------------------------------------------------------------
 
 -- | Run an `ST` action in a `ParserST`.
@@ -414,26 +432,81 @@ unsafeSpanToByteString (Span l r) =
 
 --------------------------------------------------------------------------------
 
+-- | Isolate the given parser up to (excluding) the next null byte.
+--
+-- Like 'isolate', all isolated bytes must be consumed. The null byte is
+-- consumed afterwards.
+--
+-- Useful for defining parsers for null-terminated data.
+isolateToNextNull :: ParserT st r e a -> ParserT st r e a
+isolateToNextNull (ParserT p) = ParserT \fp !r eob s n st -> go fp r n eob s st s
+  where
+    goP fp r n sNull s0 st =
+        case p fp r sNull s0 n st of
+          OK# st' a s' n' ->
+            case eqAddr# s' sNull of
+              1# -> -- consumed up to null: skip null, return
+                    OK#   st' a (sNull `plusAddr#` 1#) n'
+              _  -> Fail# st' -- didn't consume fully up to null: fail
+          x -> x
+
+    go8 fp r n eob s0 st s =
+        case eqAddr# eob s of
+          1# -> Fail# st -- end of input, no null: fail
+          _  ->
+            let s' = s `plusAddr#` 1# in
+#if MIN_VERSION_base(4,16,0)
+            -- below may be made clearer with ExtendedLiterals (GHC 9.8)
+            case eqWord8# (indexWord8OffAddr# s 0#) (wordToWord8# 0##) of
+#else
+            case eqWord# (indexWord8OffAddr# s 0#) 0## of
+#endif
+              1# -> goP fp r n s   s0 st    --     0x00: isolate, execute parser
+              _  -> go8 fp r n eob s0 st s' -- not 0x00: next please!
+
+    {- The "find first null byte" algorithms used here are adapted from
+       Hacker's Delight (2012) ch.6.
+
+    We read a word (8 bytes) at a time for efficiency. The internal algorithm
+    does byte indexing, thus endianness matters. We switch between indexing
+    algorithms depending on compile-time native endianness. (The code
+    surrounding the indexing is endian-independent, so we do this inline).
+    -}
+    go fp r n eob s0 st s =
+        let sWord = s `plusAddr#` 8# in
+        case gtAddr# sWord eob of
+          1# -> -- <  8 bytes of input: revert to scanning byte by byte
+            -- we _could_ operate on a word and simply ensure not to use the
+            -- out-of-bounds data, which would be faster, but the act of
+            -- reading could probably segfault
+            go8 fp r n eob s0 st s
+          _  -> -- >= 8 bytes of input: use efficient 8-byte scanning
+#if defined(WORDS_BIGENDIAN)
+            -- big-endian ("L->R"): find leftmost null byte
+            let !x@(I# x#) = Common.zbytel'intermediate (I# (indexIntOffAddr# s 0#)) in
+#else
+            -- little-endian ("R->L"): find rightmost null byte
+            let !x@(I# x#) = Common.zbyter'intermediate (I# (indexIntOffAddr# s 0#)) in
+#endif
+            case x# ==# 0# of
+              1# -> go fp r n eob s0 st sWord -- no 0x00 in next word
+              _  -> -- 0x00 somewhere in next word
+#if defined(WORDS_BIGENDIAN)
+                let !(I# nullIdx#) = Common.zbytel'toIdx x in
+#else
+                let !(I# nullIdx#) = Common.zbyter'toIdx x in
+                -- TO TEST BE ON LE: change above CPP to zbytel, uncomment below
+                -- let !(I# nullIdx#) = Common.zbytel'toIdx (I# (word2Int# (byteSwap# (int2Word# x#)))) in
+#endif
+                let sNull = s `plusAddr#` nullIdx# in
+                goP fp r n sNull s0 st
+{-# inline isolateToNextNull #-}
+
 -- | Read a null-terminated bytestring (a C-style string).
 --
 -- Consumes the null terminator.
 anyCString :: ParserT st r e B.ByteString
-anyCString = ParserT \fp !r eob s n st -> go' fp eob s n st
-  where
-    go' fp eob s0 n st = go 0# s0 n
-      where
-        go n# s n = case eqAddr# eob s of
-          1# -> Fail# st
-          _  ->
-            let s' = plusAddr# s 1#
-#if MIN_VERSION_base(4,16,0)
-            -- TODO below is a candidate for improving with ExtendedLiterals!
-            in  case eqWord8# (indexWord8OffAddr# s 0#) (wordToWord8# 0##) of
-#else
-            in  case eqWord# (indexWord8OffAddr# s 0#) 0## of
-#endif
-                  1# -> OK# st (B.PS (ForeignPtr s0 fp) 0 (I# n#)) s' n
-                  _  -> go (n# +# 1#) s' n
+anyCString = isolateToNextNull takeRest
 {-# inline anyCString #-}
 
 -- | Read a null-terminated bytestring (a C-style string), where the bytestring
